@@ -6,19 +6,126 @@ pub mod error;
 pub mod mwaf;
 pub mod uvfits;
 
-use std::path::Path;
+use std::{ops::Range, path::Path};
 
-use cxx::UniquePtr;
 use log::trace;
-use mwalib::CorrelatorContext;
+use mwa_rust_core::{mwalib::CorrelatorContext, Jones, LatLngHeight};
+use ndarray::{Array3, ArrayView3, ArrayViewMut3};
 use uvfits::UvfitsWriter;
 
-use crate::{pos::earth::LatLngHeight, CxxFlagMask, CxxImageSet};
+use self::error::IOError;
 
-use self::error::UvfitsWriteError;
+/// The container has visibilities which can be read by passing in a mwalib
+/// context and the range of values to read.
+pub trait ReadableVis: Sync + Send {
+    /// Read the visibilities and weights for the selected timesteps, coarse
+    /// channels and baselines into the provided arrays.
+    ///
+    /// # Errors
+    ///
+    /// Can throw IOError if there is an issue reading.
+    ///
+    /// TODO: reduce number of arguments.
+    #[allow(clippy::too_many_arguments)]
+    fn read_vis_mwalib(
+        &self,
+        jones_array: ArrayViewMut3<Jones<f32>>,
+        weight_array: ArrayViewMut3<f32>,
+        context: &CorrelatorContext,
+        timestep_range: &Range<usize>,
+        coarse_chan_range: &Range<usize>,
+        baseline_idxs: &[usize],
+    ) -> Result<(), IOError>;
+}
 
-/// Write the given [`BirliContext`] to a uvfits file.
+/// The container can accept visibilities by passing in the range of mwalib
+/// indices corresponding to the visibilities being written.
+pub trait WriteableVis: Sync + Send {
+    /// Write visibilities and weights from the arrays. Timestep, coarse channel
+    /// and baseline indices are needed for labelling the visibility
+    ///
+    /// # Errors
+    ///
+    /// Can throw IOError if there is an issue writing.
+    ///
+    /// TODO: reduce number of arguments.
+    #[allow(clippy::too_many_arguments)]
+    fn write_vis_mwalib(
+        &mut self,
+        jones_array: ArrayView3<Jones<f32>>,
+        weight_array: ArrayView3<f32>,
+        context: &CorrelatorContext,
+        timestep_range: &Range<usize>,
+        coarse_chan_range: &Range<usize>,
+        baseline_idxs: &[usize],
+    ) -> Result<(), IOError>;
+}
+
+/// Write the given ndarrays of flags and [`Jones`] matrix visibilities to a
+/// uvfits file.
 ///
+/// mwalib timestep, coarse channel and baseline indices are needed to map between
+/// indices in the arrays and indices according to mwalib, which are not the same.
+///
+/// # Examples
+///
+/// ```rust
+/// use tempfile::tempdir;
+/// use birli::{
+///     get_flaggable_timesteps,
+///     context_to_jones_array,
+///     write_uvfits,
+///     mwa_rust_core::mwalib::CorrelatorContext
+/// };
+///
+/// // define our input files
+/// let metafits_path = "tests/data/1196175296_mwa_ord/1196175296.metafits";
+/// let gpufits_paths = vec![
+///     "tests/data/1196175296_mwa_ord/1196175296_20171201145440_gpubox01_00.fits",
+/// ];
+///
+/// // define a temporary directory for output files
+/// let tmp_dir = tempdir().unwrap();
+///
+/// // define our output flag file template
+/// let uvfits_out = tmp_dir.path().join("synthetic.uvfits");
+///
+/// // Create an mwalib::CorrelatorContext for accessing visibilities.
+/// let context = CorrelatorContext::new(&metafits_path, &gpufits_paths).unwrap();
+///
+/// // Determine which timesteps and coarse channels we want to use
+/// let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+/// let img_timestep_range =
+///     *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
+/// let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
+/// let img_coarse_chan_range =
+///     *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+///
+///
+/// // generate an array of jones matrices
+/// let (jones_array, flag_array) = context_to_jones_array(
+///     &context,
+///     &img_timestep_range,
+///     &img_coarse_chan_range,
+///     None,
+/// );
+///
+/// // write the visibilities to disk as .uvfits
+///
+/// let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
+///
+/// write_uvfits(
+///     uvfits_out.as_path(),
+///     &context,
+///     &jones_array,
+///     &flag_array,
+///     &img_timestep_range,
+///     &img_coarse_chan_range,
+///     &baseline_idxs,
+///     None,
+/// )
+/// .unwrap();
+/// ```
 /// # Errors
 ///
 /// See: [`UvfitsWriter`]
@@ -28,34 +135,31 @@ use self::error::UvfitsWriteError;
 pub fn write_uvfits<'a>(
     filename: &'a Path,
     context: &CorrelatorContext,
-    baseline_idxs: &[usize],
-    baseline_imgsets: &[UniquePtr<CxxImageSet>],
-    baseline_flagmasks: &[UniquePtr<CxxFlagMask>],
-    img_timestep_idxs: &[usize],
-    img_coarse_chan_idxs: &[usize],
+    jones_array: &Array3<Jones<f32>>,
+    flag_array: &Array3<bool>,
+    mwalib_timestep_range: &Range<usize>,
+    mwalib_coarse_chan_range: &Range<usize>,
+    mwalib_baseline_idxs: &[usize],
     array_pos: Option<LatLngHeight>,
-) -> Result<(), UvfitsWriteError> {
+) -> Result<(), IOError> {
     trace!("start write_uvfits");
 
     let mut uvfits_writer = UvfitsWriter::from_mwalib(
         filename,
         context,
-        img_timestep_idxs,
-        img_coarse_chan_idxs,
-        baseline_idxs,
+        mwalib_timestep_range,
+        mwalib_coarse_chan_range,
+        mwalib_baseline_idxs,
         array_pos,
     )?;
 
-    let mut fits_file = uvfits_writer.open().unwrap();
-
-    uvfits_writer.write_baseline_imgset_flagmasks(
-        &mut fits_file,
+    uvfits_writer.write_jones_flags(
         context,
-        baseline_idxs,
-        baseline_imgsets,
-        baseline_flagmasks,
-        img_timestep_idxs,
-        img_coarse_chan_idxs,
+        jones_array,
+        flag_array,
+        mwalib_timestep_range,
+        mwalib_coarse_chan_range,
+        mwalib_baseline_idxs,
     )?;
 
     uvfits_writer.write_ants_from_mwalib(&context.metafits_context)?;
@@ -66,26 +170,28 @@ pub fn write_uvfits<'a>(
 }
 
 #[cfg(test)]
-mod tests {
-
+#[cfg(feature = "aoflagger")]
+/// Tests which require the use of the aoflagger feature
+mod tests_aoflagger {
+    use crate::{
+        context_to_jones_array, flags::flag_jones_array_existing, get_antenna_flags,
+        get_flaggable_timesteps, init_flag_array, write_uvfits,
+    };
+    use aoflagger_sys::cxx_aoflagger_new;
     use fitsio::errors::check_status as fits_check_status;
     use float_cmp::{approx_eq, F32Margin};
     use itertools::izip;
-    use mwalib::{
-        CorrelatorContext, _get_required_fits_key, _open_fits, _open_hdu, fits_open, fits_open_hdu,
-        get_required_fits_key,
+    use mwa_rust_core::{
+        fitsio, fitsio_sys,
+        mwalib::{
+            _get_required_fits_key, _open_fits, _open_hdu, fits_open, fits_open_hdu,
+            get_required_fits_key, CorrelatorContext,
+        },
     };
     use tempfile::tempdir;
 
-    use crate::{
-        context_to_baseline_imgsets, cxx_aoflagger_new, flag_imgsets_existing, get_antenna_flags,
-        get_flaggable_timesteps, init_baseline_flagmasks,
-    };
-
-    use super::write_uvfits;
-
     #[test]
-    fn write_uvfits_mwa_ord() {
+    fn write_uvfits_flags() {
         // define our input files
         let metafits_path = "tests/data/1196175296_mwa_ord/1196175296.metafits";
         let gpufits_paths = vec![
@@ -98,7 +204,7 @@ mod tests {
         // define a temporary directory for output files
         let tmp_dir = tempdir().unwrap();
 
-        // define our output flag file template
+        // define our output uvfits path
         let uvfits_out = tmp_dir.path().join("1297526432.uvfits");
 
         // Create an mwalib::CorrelatorContext for accessing visibilities.
@@ -108,50 +214,53 @@ mod tests {
         let aoflagger = unsafe { cxx_aoflagger_new() };
 
         // Determine which timesteps and coarse channels we want to use
-        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
         let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        assert_eq!(img_timestep_idxs.len(), 4);
+        let img_timestep_range =
+            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
+        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let img_coarse_chan_range =
+            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
 
         // Prepare our flagmasks with known bad antennae
-        let mut baseline_flagmasks = init_baseline_flagmasks(
-            &aoflagger,
+        let flag_array = init_flag_array(
             &context,
-            img_coarse_chan_idxs,
-            &img_timestep_idxs,
+            &img_timestep_range,
+            &img_coarse_chan_range,
             Some(get_antenna_flags(&context)),
         );
 
-        // generate imagesets for each baseline in the format required by aoflagger
-        let baseline_imgsets = context_to_baseline_imgsets(
-            &aoflagger,
+        // generate an array of jones matrices
+        let (jones_array, flag_array) = context_to_jones_array(
             &context,
-            img_coarse_chan_idxs,
-            &img_timestep_idxs,
-            Some(&mut baseline_flagmasks),
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            Some(flag_array),
         );
 
         // use the default strategy file location for MWA
         let strategy_filename = &aoflagger.FindStrategyFileMWA();
 
         // run the strategy on the imagesets, and get the resulting flagmasks for each baseline
-        flag_imgsets_existing(
+        let flag_array = flag_jones_array_existing(
             &aoflagger,
             strategy_filename,
-            &baseline_imgsets,
-            &mut baseline_flagmasks,
+            &jones_array,
+            Some(flag_array),
             true,
         );
 
         // write the visibilities to disk as .uvfits
 
-        let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
         write_uvfits(
             uvfits_out.as_path(),
             &context,
+            &jones_array,
+            &flag_array,
+            &img_timestep_range,
+            &img_coarse_chan_range,
             &baseline_idxs,
-            &baseline_imgsets,
-            &baseline_flagmasks,
-            &img_timestep_idxs,
-            img_coarse_chan_idxs,
             None,
         )
         .unwrap();
