@@ -6,32 +6,33 @@
 //! Most of this was blatently stolen (with permission) from [Chris Jordan](https://github.com/cjordan)
 
 // use std::collections::HashSet;
-use std::ffi::CString;
 use std::path::Path;
+use std::{ffi::CString, ops::Range};
 
-use erfa_sys::ERFA_DJM0;
 use fitsio::{errors::check_status as fits_check_status, FitsFile};
-use hifitime::Epoch;
 use indicatif::ProgressStyle;
 use itertools::izip;
 use log::warn;
-// use ndarray::prelude::*;
+use mwa_rust_core::{
+    constants::VEL_C,
+    erfa_sys,
+    erfa_sys::ERFA_DJM0,
+    fitsio, fitsio_sys,
+    hifitime::Epoch,
+    mwalib::{CorrelatorContext, MetafitsContext},
+    precession::*,
+    time, Jones, LatLngHeight, RADec, XyzGeodetic, ENH, UVW,
+};
+use ndarray::{Array3, ArrayView3, Axis};
 
-use super::error::UvfitsWriteError;
-use crate::{
-    constants::HIFITIME_GPS_FACTOR,
-    cxx_aoflagger::ffi::{CxxFlagMask, CxxImageSet},
-    pos::{earth::LatLngHeight, precess::precess_time, RADec, XyzGeodetic, ENH, UVW},
-};
-use cxx::UniquePtr;
-use mwalib::{
-    fitsio, fitsio_sys, Baseline, CorrelatorContext, MetafitsContext, MWA_ALTITUDE_METRES,
-    MWA_LATITUDE_RADIANS, MWA_LONGITUDE_RADIANS, SPEED_OF_LIGHT_IN_VACUUM_M_PER_S,
-};
+use crate::flags::flag_to_weight_array;
+
+use super::error::{IOError, UvfitsWriteError};
+use super::WriteableVis;
 
 /// From a `hifitime` [Epoch], get a formatted date string with the hours,
 /// minutes and seconds set to 0.
-fn get_truncated_date_string(epoch: &Epoch) -> String {
+fn get_truncated_date_string(epoch: Epoch) -> String {
     let (year, month, day, _, _, _, _) = epoch.as_gregorian_utc();
     format!(
         "{year}-{month:02}-{day:02}T00:00:00.0",
@@ -120,8 +121,9 @@ impl<'a> UvfitsWriter<'a> {
     /// `start_epoch` can be calculated from GPS Time using the hifitime library, e.g.
     ///
     /// ```rust
+    /// use mwa_rust_core::time;
     /// let first_gps_time = 1196175296.0;
-    /// let start_epoch = hifitime::Epoch::from_tai_seconds(first_gps_time + 19.0 + birli::constants::HIFITIME_GPS_FACTOR);
+    /// let start_epoch = time::gps_to_epoch(first_gps_time);
     /// ```
     ///
     /// `centre_freq_hz` is the centre frequency of the coarse band that this
@@ -222,7 +224,7 @@ impl<'a> UvfitsWriter<'a> {
                 )?;
             }
         }
-        hdu.write_key(&mut u, "DATE-OBS", get_truncated_date_string(&start_epoch))?;
+        hdu.write_key(&mut u, "DATE-OBS", get_truncated_date_string(start_epoch))?;
 
         // Dimensions.
         hdu.write_key(&mut u, "CTYPE2", "COMPLEX")?;
@@ -303,11 +305,7 @@ impl<'a> UvfitsWriter<'a> {
                 // This is at least partly due to different constants (the altitude is
                 // definitely slightly different), but possibly also because ERFA is
                 // more accurate than cotter's "homebrewed" Geodetic2XYZ.
-                LatLngHeight {
-                    longitude_rad: MWA_LONGITUDE_RADIANS,
-                    latitude_rad: MWA_LATITUDE_RADIANS,
-                    height_metres: MWA_ALTITUDE_METRES,
-                }
+                LatLngHeight::new_mwa()
             }
         };
 
@@ -324,8 +322,10 @@ impl<'a> UvfitsWriter<'a> {
 
     /// Create a new uvfits file at the specified filename using an [`mwalib::CorrelatorContext`]
     ///
-    /// start epoch is determined by the first time provided in `img_timestep_idxs`
-    /// which may not necessarily match the obsid.
+    /// # Details
+    ///
+    /// start epoch is determined by `mwalib_timestep_range.start` which may not necessarily match
+    /// the obsid.
     ///
     /// # Errors
     ///
@@ -333,22 +333,23 @@ impl<'a> UvfitsWriter<'a> {
     pub fn from_mwalib(
         filename: &'a Path,
         context: &CorrelatorContext,
-        img_timestep_idxs: &[usize],
-        img_coarse_chan_idxs: &[usize],
-        img_baseline_idxs: &[usize],
+        mwalib_timestep_range: &Range<usize>,
+        mwalib_coarse_chan_range: &Range<usize>,
+        mwalib_baseline_idxs: &[usize],
         array_pos: Option<LatLngHeight>,
     ) -> Result<Self, UvfitsWriteError> {
-        let num_fine_chans_per_coarse = context.metafits_context.num_corr_fine_chans_per_coarse;
-        let num_img_coarse_chans = img_coarse_chan_idxs.len();
-        let num_img_chans = num_fine_chans_per_coarse * num_img_coarse_chans;
-        let first_gps_time = context.timesteps[img_timestep_idxs[0]].gps_time_ms as f64 / 1000.0;
-        let start_epoch = Epoch::from_tai_seconds(first_gps_time + 19.0 + HIFITIME_GPS_FACTOR);
+        let fine_chans_per_coarse = context.metafits_context.num_corr_fine_chans_per_coarse;
+        let num_img_coarse_chans = mwalib_coarse_chan_range.len();
+        let num_img_chans = fine_chans_per_coarse * num_img_coarse_chans;
+        let first_gps_time =
+            context.timesteps[mwalib_timestep_range.start].gps_time_ms as f64 / 1000.0;
+        let start_epoch = time::gps_to_epoch(first_gps_time);
         let phase_centre = RADec::from_mwalib_phase_or_pointing(&context.metafits_context);
         let centre_freq_chan = num_img_chans / 2;
         Self::new(
             filename,
-            img_timestep_idxs.len(),
-            img_baseline_idxs.len(),
+            mwalib_timestep_range.len(),
+            mwalib_baseline_idxs.len(),
             num_img_chans,
             start_epoch,
             context.metafits_context.corr_fine_chan_width_hz as f64,
@@ -458,7 +459,7 @@ impl<'a> UvfitsWriter<'a> {
         hdu.write_key(&mut uvfits, "GSTIA0", gst)?;
         hdu.write_key(&mut uvfits, "DEGPDY", 3.60985e2)?; // Earth's rotation rate
 
-        let date_truncated = get_truncated_date_string(&self.start_epoch);
+        let date_truncated = get_truncated_date_string(self.start_epoch);
         hdu.write_key(&mut uvfits, "RDATE", date_truncated)?;
 
         hdu.write_key(&mut uvfits, "POLARX", 0.0)?;
@@ -655,17 +656,15 @@ impl<'a> UvfitsWriter<'a> {
     ///
     /// TODO: Assumes that all fine channels are written in `vis.` This needs to
     /// be updated to add visibilities to an existing uvfits row.
-    ///
-    /// TODO: replace all these args with birli_context
     #[allow(clippy::too_many_arguments)]
     #[inline]
-    pub fn write_vis(
+    pub fn write_vis_row(
         &mut self,
         uvfits: &mut FitsFile,
-        uvw: &UVW,
+        uvw: UVW,
         tile_index1: usize,
         tile_index2: usize,
-        epoch: &Epoch,
+        epoch: Epoch,
         vis: &[f32],
     ) -> Result<(), UvfitsWriteError> {
         if self.current_num_rows + 1 > self.total_num_rows {
@@ -676,9 +675,9 @@ impl<'a> UvfitsWriter<'a> {
         }
 
         let mut row = Vec::with_capacity(5 + vis.len());
-        row.push((uvw.u / SPEED_OF_LIGHT_IN_VACUUM_M_PER_S) as f32);
-        row.push((uvw.v / SPEED_OF_LIGHT_IN_VACUUM_M_PER_S) as f32);
-        row.push((uvw.w / SPEED_OF_LIGHT_IN_VACUUM_M_PER_S) as f32);
+        row.push((uvw.u / VEL_C) as f32);
+        row.push((uvw.v / VEL_C) as f32);
+        row.push((uvw.w / VEL_C) as f32);
         row.push(encode_uvfits_baseline(tile_index1 + 1, tile_index2 + 1) as f32);
         let jd_trunc = self.start_epoch.as_jde_utc_days().floor() + 0.5;
         let jd_frac = epoch.as_jde_utc_days() - jd_trunc;
@@ -703,9 +702,11 @@ impl<'a> UvfitsWriter<'a> {
         Ok(())
     }
 
-    /// Write visibilty rows into the uvfits file from the provided
+    /// Write visibilty and weight rows into the uvfits file from the provided
     /// [`mwalib::CorrelatorContext`].
     ///
+    /// # Details
+    ///
     /// `uvfits` must have been opened in write mode and currently have HDU 0
     /// open. The [FitsFile] must be supplied to this function to force the
     /// caller to think about calling this function efficiently; opening the
@@ -715,17 +716,20 @@ impl<'a> UvfitsWriter<'a> {
     /// `baseline_idxs` the baseline indices (according to mwalib)
     /// which should be written to the file
     ///
-    /// `baseline_imgsets` a [`CxxImageSet`] for each baseline in `baseline_idxs`
+    /// `jones_array` a [`ndarray::Array3`] of [`Jones`] visibilities with dimensions
+    /// [timestep][channel][baselines]
     ///
-    /// `baseline_flagmasks` a [`CxxFlagMask`] for each baseline in `baseline_idxs`
+    /// `flag_array` a [`ndarray::Array3`] of boolean flags with dimensions
+    /// identical dimensions to `jones_array`
     ///
-    /// `img_timestep_idxs` the timestep indices (according to mwalib)
-    /// which are used in all of the images in `baseline_imgsets` and all of the
-    /// flagmasks in `baseline_flagmasks`
+    /// `mwalib_timestep_range` the range of timestep indices (according to mwalib)
+    /// which are used in the visibility and flag arrays
     ///
-    /// `img_coarse_chan_idxs` the coarse channel indices (according to mwalib)
-    /// which are used in all of the images in `baseline_imgsets` and all of the
-    /// flagmasks in `baseline_flagmasks`
+    /// `mwalib_coarse_chan_range` the range of coarse channel indices (according to mwalib)
+    /// which are used in the visibility and flag arrays
+    ///
+    /// `mwalib_baseline_idxs` the baseline indices (according to mwalib) used
+    /// in the visibility and flag arrays
     ///
     /// TODO: handle averaging
     ///
@@ -736,47 +740,82 @@ impl<'a> UvfitsWriter<'a> {
     ///
     /// TODO: replace all these args with birli_context
     #[allow(clippy::too_many_arguments)]
-    pub fn write_baseline_imgset_flagmasks(
+    pub fn write_jones_flags(
         &mut self,
-        uvfits: &mut FitsFile,
         context: &CorrelatorContext,
+        jones_array: &Array3<Jones<f32>>,
+        flag_array: &Array3<bool>,
+        mwalib_timestep_range: &Range<usize>,
+        mwalib_coarse_chan_range: &Range<usize>,
+        mwalib_baseline_idxs: &[usize],
+    ) -> Result<(), IOError> {
+        let weight_array = flag_to_weight_array(context, flag_array.view());
+        self.write_vis_mwalib(
+            jones_array.view(),
+            weight_array.view(),
+            context,
+            mwalib_timestep_range,
+            mwalib_coarse_chan_range,
+            mwalib_baseline_idxs,
+        )
+    }
+}
+
+// In order to get rid of context, would need to :
+// - replace timestep range with slice of epochs
+// - replace coarse chan range with...
+// - replace baseline indices with ...
+// - pass in tiles xyz geodetic
+impl<'a> WriteableVis for UvfitsWriter<'a> {
+    fn write_vis_mwalib(
+        &mut self,
+        jones_array: ArrayView3<Jones<f32>>,
+        weight_array: ArrayView3<f32>,
+        context: &CorrelatorContext,
+        timestep_range: &Range<usize>,
+        coarse_chan_range: &Range<usize>,
         baseline_idxs: &[usize],
-        baseline_imgsets: &[UniquePtr<CxxImageSet>],
-        baseline_flagmasks: &[UniquePtr<CxxFlagMask>],
-        img_timestep_idxs: &[usize],
-        img_coarse_chan_idxs: &[usize],
-    ) -> Result<(), UvfitsWriteError> {
+    ) -> Result<(), IOError> {
+        let mut uvfits = self.open()?;
+
+        let fine_chans_per_coarse = context.metafits_context.num_corr_fine_chans_per_coarse;
+
+        let jones_dims = jones_array.dim();
+        let weight_dims = weight_array.dim();
+        assert_eq!(jones_dims, weight_dims);
+
+        let num_img_timesteps = timestep_range.len();
+        assert_eq!(num_img_timesteps, jones_dims.0);
+
+        let num_img_coarse_chans = coarse_chan_range.len();
+        let num_img_chans = fine_chans_per_coarse * num_img_coarse_chans;
+        assert_eq!(num_img_coarse_chans * fine_chans_per_coarse, jones_dims.1);
+
         let num_baselines = baseline_idxs.len();
-        assert_eq!(num_baselines, baseline_imgsets.len());
-        assert_eq!(num_baselines, baseline_flagmasks.len());
+        assert_eq!(num_baselines, jones_dims.2);
 
-        let tiles_xyz_geod = XyzGeodetic::get_tiles_mwalib(&context.metafits_context);
-
-        let num_fine_chans_per_coarse = context.metafits_context.num_corr_fine_chans_per_coarse;
-        let num_img_coarse_chans = img_coarse_chan_idxs.len();
-        let num_img_chans = num_fine_chans_per_coarse * num_img_coarse_chans;
-
-        let num_img_timesteps = img_timestep_idxs.len();
         assert_eq!(self.total_num_rows, num_img_timesteps * num_baselines);
 
+        let tiles_xyz_geod = XyzGeodetic::get_tiles_mwa(&context.metafits_context);
+
+        let img_timesteps = &context.timesteps[timestep_range.clone()];
+        let img_baselines = baseline_idxs
+            .iter()
+            .map(|&idx| {
+                context
+                    .metafits_context
+                    .baselines
+                    .get(idx)
+                    .unwrap()
+                    .to_owned()
+            })
+            .collect::<Vec<_>>();
+
         // Weights are normalized so that default res of 10 kHz, 1s has weight of "1" per sample
-
-        // TODO: deal with weight factor when doing averaging.
-        // TODO: deal with passband gains
         let integration_time_s = context.metafits_context.corr_int_time_ms as f64 / 1000.0;
-        let fine_chan_width_hz = context.metafits_context.corr_fine_chan_width_hz as f64;
-        let weight_factor = fine_chan_width_hz * integration_time_s / 10000.0;
-
-        // MWA/CASA/AOFlagger visibility order is XX,XY,YX,YY
-        // UVFits visibility order is XX,YY,XY,YX
-        let pol_order = vec![0, 3, 1, 2];
-        // Real and imaginary component offsets within image width
-        // let re_im = vec![0, 1];
-
-        let num_blts = img_timestep_idxs.len() * baseline_idxs.len();
 
         // Create a progress bar to show the writing status
-        let write_progress = indicatif::ProgressBar::new(num_blts as u64);
+        let write_progress = indicatif::ProgressBar::new(self.total_num_rows as u64);
         write_progress.set_style(
             ProgressStyle::default_bar()
                 .template(
@@ -786,70 +825,62 @@ impl<'a> UvfitsWriter<'a> {
         );
         write_progress.set_message("write uv vis");
 
-        for (img_timestep_idx, &timestep_idx) in img_timestep_idxs.iter().enumerate() {
-            let gps_time_s = context.timesteps[timestep_idx].gps_time_ms as f64 / 1000.0;
-            // TODO: document where this 19 comes from?
-            let epoch = Epoch::from_tai_seconds(
-                gps_time_s + 19.0 + HIFITIME_GPS_FACTOR + integration_time_s / 2.0,
-            );
+        let mut vis: Vec<f32> = vec![0.0; 12 * num_img_chans];
+
+        for (timestep, jones_timestep_view, weight_timestep_view) in izip!(
+            img_timesteps.iter(),
+            jones_array.outer_iter(),
+            weight_array.outer_iter()
+        ) {
+            let gps_time_s = timestep.gps_time_ms as f64 / 1000.0;
+            let epoch = time::gps_to_epoch(gps_time_s + integration_time_s / 2.0);
 
             let prec_info = precess_time(
-                &self.phase_centre,
-                &epoch,
+                self.phase_centre,
+                epoch,
                 self.array_pos.longitude_rad,
                 self.array_pos.latitude_rad,
             );
 
             let tiles_xyz_precessed = prec_info.precess_xyz_parallel(&tiles_xyz_geod);
 
-            for (&baseline_idx, imgset, flagmask) in
-                izip!(baseline_idxs, baseline_imgsets, baseline_flagmasks)
-            {
-                let baseline: &Baseline = &context.metafits_context.baselines[baseline_idx];
-                let imgset: &UniquePtr<CxxImageSet> = imgset;
-                let img_stride: usize = imgset.HorizontalStride();
-
-                let flagmask: &UniquePtr<CxxFlagMask> = flagmask;
-                let flag_buffer: &[bool] = flagmask.Buffer();
-                let flag_stride: usize = flagmask.HorizontalStride();
-
+            for (baseline, jones_baseline_view, weight_baseline_view) in izip!(
+                img_baselines.iter(),
+                jones_timestep_view.axis_iter(Axis(1)),
+                weight_timestep_view.axis_iter(Axis(1))
+            ) {
                 let ant1_idx = baseline.ant1_index;
                 let ant2_idx = baseline.ant2_index;
 
                 let baseline_xyz_precessed =
-                    tiles_xyz_precessed[ant1_idx].clone() - &tiles_xyz_precessed[ant2_idx];
-                let uvw = UVW::from_xyz(&baseline_xyz_precessed, &prec_info.hadec_j2000);
+                    tiles_xyz_precessed[ant1_idx] - tiles_xyz_precessed[ant2_idx];
+                let uvw = UVW::from_xyz(baseline_xyz_precessed, prec_info.hadec_j2000);
 
-                // TODO: this is extremely inefficient.
+                // MWA/CASA/AOFlagger visibility order is XX,XY,YX,YY
+                // UVFits visibility order is XX,YY,XY,YX
 
-                let vis: Vec<f32> = (0..num_img_chans)
-                    .flat_map(|chan_idx| {
-                        pol_order
-                            .iter()
-                            .flat_map(|img_pol_idx| {
-                                let img_buffer_re: &[f32] = imgset.ImageBuffer(img_pol_idx * 2);
-                                let img_buffer_im: &[f32] = imgset.ImageBuffer(img_pol_idx * 2 + 1);
-                                let vis_re =
-                                    img_buffer_re[chan_idx * img_stride + img_timestep_idx];
-                                let vis_im =
-                                    img_buffer_im[chan_idx * img_stride + img_timestep_idx];
-                                let flag = flag_buffer[chan_idx * flag_stride + img_timestep_idx];
-                                // TODO: weight from flags
-                                let mut weight = weight_factor as f32;
-                                if flag {
-                                    weight *= -1.0;
-                                }
-                                vec![vis_re, vis_im, weight]
-                            })
-                            .collect::<Vec<_>>()
-                    })
-                    .collect();
+                izip!(
+                    jones_baseline_view.iter(),
+                    weight_baseline_view.iter(),
+                    vis.chunks_exact_mut(12)
+                )
+                .for_each(|(jones, &weight, vis_chunk)| {
+                    // TODO: something a little prettier
+                    vis_chunk[0] = jones[0].re;
+                    vis_chunk[1] = jones[0].im;
+                    vis_chunk[2] = weight;
+                    vis_chunk[3] = jones[3].re;
+                    vis_chunk[4] = jones[3].im;
+                    vis_chunk[5] = weight;
+                    vis_chunk[6] = jones[1].re;
+                    vis_chunk[7] = jones[1].im;
+                    vis_chunk[8] = weight;
+                    vis_chunk[9] = jones[2].re;
+                    vis_chunk[10] = jones[2].im;
+                    vis_chunk[11] = weight;
+                });
 
-                // TODO: calculate weights
-                let result = self.write_vis(uvfits, &uvw, ant1_idx, ant2_idx, &epoch, &vis.clone());
-                if let Err(err) = result {
-                    return Err(err);
-                };
+                self.write_vis_row(&mut uvfits, uvw, ant1_idx, ant2_idx, epoch, &vis)?;
 
                 write_progress.inc(1);
             }
@@ -863,30 +894,10 @@ impl<'a> UvfitsWriter<'a> {
 
 #[cfg(test)]
 mod tests {
-
     use super::*;
-    use mwalib::{
-        _get_fits_col, _get_required_fits_key, _open_fits, _open_hdu, fits_open, fits_open_hdu,
-        get_fits_col, get_required_fits_key,
-    };
+    use mwa_rust_core::time;
     use tempfile::NamedTempFile;
 
-    use float_cmp::{approx_eq, F32Margin, F64Margin};
-
-    use crate::{
-        constants::{
-            COTTER_MWA_HEIGHT_METRES, COTTER_MWA_LATITUDE_RADIANS, COTTER_MWA_LONGITUDE_RADIANS,
-            HIFITIME_GPS_FACTOR,
-        },
-        context_to_baseline_imgsets, cxx_aoflagger_new, flag_imgsets_existing, get_antenna_flags,
-        get_flaggable_timesteps, init_baseline_flagmasks,
-    };
-
-    use fitsio::{
-        errors::check_status as fits_check_status,
-        hdu::{FitsHdu, HduInfo},
-    };
-
     #[test]
     // Make a tiny uvfits file. The result has been verified by CASA's
     // "importuvfits" function.
@@ -896,7 +907,7 @@ mod tests {
         let num_baselines = 3;
         let num_chans = 2;
         let obsid = 1065880128;
-        let start_epoch = Epoch::from_tai_seconds(obsid as f64 + 19.0 + HIFITIME_GPS_FACTOR);
+        let start_epoch = time::gps_to_epoch(obsid as f64);
 
         let mut u = UvfitsWriter::new(
             tmp_uvfits_file.path(),
@@ -923,12 +934,12 @@ mod tests {
                     _ => unreachable!(),
                 };
 
-                u.write_vis(
+                u.write_vis_row(
                     &mut f,
-                    &UVW::default(),
+                    UVW::default(),
                     tile1,
                     tile2,
-                    &start_epoch,
+                    start_epoch,
                     (baseline_index..baseline_index + num_chans)
                         .into_iter()
                         .map(|int| int as f32)
@@ -950,6 +961,37 @@ mod tests {
             .collect();
         u.write_uvfits_antenna_table(&names, &positions).unwrap();
     }
+}
+#[cfg(test)]
+#[cfg(feature = "aoflagger")]
+/// Tests which require the use of the aoflagger feature
+mod tests_aoflagger {
+    use super::*;
+    use mwa_rust_core::{
+        constants::{
+            COTTER_MWA_HEIGHT_METRES, COTTER_MWA_LATITUDE_RADIANS, COTTER_MWA_LONGITUDE_RADIANS,
+        },
+        mwalib::{
+            _get_fits_col, _get_required_fits_key, _open_fits, _open_hdu, fits_open, fits_open_hdu,
+            get_fits_col, get_required_fits_key, CorrelatorContext,
+        },
+    };
+    use tempfile::NamedTempFile;
+
+    use float_cmp::{approx_eq, F32Margin, F64Margin};
+
+    use crate::get_flaggable_timesteps;
+
+    use fitsio::{
+        errors::check_status as fits_check_status,
+        hdu::{FitsHdu, HduInfo},
+    };
+
+    use crate::{
+        context_to_jones_array, flags::flag_jones_array_existing, get_antenna_flags,
+        init_flag_array,
+    };
+    use aoflagger_sys::cxx_aoflagger_new;
 
     // TODO: dedup this from lib.rs
     fn get_mwa_ord_context() -> CorrelatorContext {
@@ -1243,7 +1285,7 @@ mod tests {
     pub fn get_group_column_description(
         fptr: &mut FitsFile,
         hdu: &FitsHdu,
-    ) -> Result<Vec<String>, mwalib::FitsError> {
+    ) -> Result<Vec<String>, crate::mwalib::FitsError> {
         let pcount: usize = get_required_fits_key!(fptr, hdu, "PCOUNT")?;
         let mut result = Vec::with_capacity(pcount);
         for p in 0..pcount {
@@ -1464,10 +1506,13 @@ mod tests {
         let tmp_uvfits_file = NamedTempFile::new().unwrap();
         // let tmp_uvfits_file = Path::new("tests/data/test_header.uvfits");
 
-        let timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        let coarse_chan_idxs = context.common_coarse_chan_indices.clone();
-
-        let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
+        let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
+        let img_timestep_range =
+            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
+        let img_coarse_chan_idxs = context.common_coarse_chan_indices.clone();
+        let img_coarse_chan_range =
+            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let img_baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
 
         let array_pos = Some(LatLngHeight {
             longitude_rad: COTTER_MWA_LONGITUDE_RADIANS,
@@ -1478,9 +1523,9 @@ mod tests {
         let u = UvfitsWriter::from_mwalib(
             tmp_uvfits_file.path(),
             &context,
-            &timestep_idxs,
-            &coarse_chan_idxs,
-            &baseline_idxs,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            &img_baseline_idxs,
             array_pos,
         )
         .unwrap();
@@ -1498,15 +1543,20 @@ mod tests {
     }
 
     #[test]
+    #[cfg(feature = "aoflagger")]
     fn uvfits_tables_from_mwalib_matches_cotter() {
         let context = get_mwa_ord_context();
 
         let tmp_uvfits_file = NamedTempFile::new().unwrap();
 
         let img_timestep_idxs = get_flaggable_timesteps(&context).unwrap();
-        let img_coarse_chan_idxs = context.common_coarse_chan_indices.clone();
-
-        let baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
+        assert_eq!(img_timestep_idxs.len(), 4);
+        let img_timestep_range =
+            *img_timestep_idxs.first().unwrap()..(*img_timestep_idxs.last().unwrap() + 1);
+        let img_coarse_chan_idxs = &context.common_coarse_chan_indices;
+        let img_coarse_chan_range =
+            *img_coarse_chan_idxs.first().unwrap()..(*img_coarse_chan_idxs.last().unwrap() + 1);
+        let img_baseline_idxs = (0..context.metafits_context.num_baselines).collect::<Vec<_>>();
 
         let array_pos = Some(LatLngHeight {
             longitude_rad: COTTER_MWA_LONGITUDE_RADIANS,
@@ -1517,58 +1567,51 @@ mod tests {
         let mut u = UvfitsWriter::from_mwalib(
             tmp_uvfits_file.path(),
             &context,
-            &img_timestep_idxs,
-            &img_coarse_chan_idxs,
-            &baseline_idxs,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            &img_baseline_idxs,
             array_pos,
         )
         .unwrap();
 
-        let mut f = u.open().unwrap();
-
         let aoflagger = unsafe { cxx_aoflagger_new() };
 
-        let mut baseline_flagmasks = init_baseline_flagmasks(
-            &aoflagger,
+        let flag_array = init_flag_array(
             &context,
-            &img_coarse_chan_idxs,
-            &img_timestep_idxs,
+            &img_timestep_range,
+            &img_coarse_chan_range,
             Some(get_antenna_flags(&context)),
         );
 
-        let baseline_imgsets = context_to_baseline_imgsets(
-            &aoflagger,
+        let (jones_array, flag_array) = context_to_jones_array(
             &context,
-            &img_coarse_chan_idxs,
-            &img_timestep_idxs,
-            None,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            Some(flag_array),
         );
 
         let strategy_filename = &aoflagger.FindStrategyFileMWA();
 
-        flag_imgsets_existing(
+        let flag_array = flag_jones_array_existing(
             &aoflagger,
             strategy_filename,
-            &baseline_imgsets,
-            &mut baseline_flagmasks,
+            &jones_array,
+            Some(flag_array),
             true,
         );
 
-        u.write_baseline_imgset_flagmasks(
-            &mut f,
+        u.write_jones_flags(
             &context,
-            &baseline_idxs,
-            &baseline_imgsets,
-            &baseline_flagmasks,
-            &img_timestep_idxs,
-            &img_coarse_chan_idxs,
+            &jones_array,
+            &flag_array,
+            &img_timestep_range,
+            &img_coarse_chan_range,
+            &img_baseline_idxs,
         )
         .unwrap();
 
         u.write_ants_from_mwalib(&context.metafits_context).unwrap();
 
-        drop(f);
-
         let cotter_uvfits_path = Path::new("tests/data/1196175296_mwa_ord/1196175296.uvfits");
 
         let mut birli_fptr = fits_open!(&tmp_uvfits_file.path()).unwrap();
